from ._recorders cimport NumpyArrayNodeRecorder, BaseConstantNodeRecorder
from pywr.parameters._parameters cimport Parameter
from .._core cimport Timestep, Scenario, ScenarioIndex


cdef class HydropowerRecorder(NumpyArrayNodeRecorder):
    cdef Parameter _water_elevation_parameter
    cdef public double turbine_elevation
    cdef public double flow_unit_conversion
    cdef public double energy_unit_conversion
    cdef public double density
    cdef public double efficiency


cdef class TotalHydroEnergyRecorder(BaseConstantNodeRecorder):
    cdef Parameter _water_elevation_parameter
    cdef public double turbine_elevation
    cdef public double flow_unit_conversion
    cdef public double energy_unit_conversion
    cdef public double density
    cdef public double efficiency


""" Generic utilities functions useful for pywr, but not specific to any of the core modules
"""
import io
import hashlib


def compute_hash(filename, algorithm="md5", chunk_size=io.DEFAULT_BUFFER_SIZE):
    """Compute the hash of a large file using hashlib"""
    h = hashlib.new(algorithm)

    with io.open(filename, mode="rb") as fh:
        for chunk in iter(lambda: fh.read(chunk_size), b""):
            h.update(chunk)

    return h.hexdigest()


class HashMismatchError(IOError):
    pass


def check_hash(filename, hash, algorithm="md5", **kwargs):
    """Check the hash for filename using the named algorithm

    If the hashes do not match a HashMismatchError error is raised.
    """

    actual_hash = compute_hash(filename, algorithm=algorithm, **kwargs)

    if hash != actual_hash:
        raise HashMismatchError(
            'Hash mismatch using {} on file: "{}"'.format(algorithm, filename)
        )



import tables
import os


class H5Store(object):
    def __init__(
        self,
        filename,
        filter_kwds=None,
        mode="r",
        title="",
        metadata=None,
        create_directories=False,
    ):
        self._opened = False
        if isinstance(filename, (str, os.PathLike)):
            # filename is a path to open
            self.filename = filename
            # Note sure how else to deal with str / unicode requirements in pytables
            # See this issue: https://github.com/PyTables/PyTables/issues/522
            import sys

            if filter_kwds:
                if sys.version_info[0] == 2 and "complib" in filter_kwds:
                    filter_kwds["complib"] = filter_kwds["complib"].encode()
                filters = tables.Filters(**filter_kwds)
            else:
                filters = None

            # Create directories for the filename if required
            if create_directories:
                try:
                    os.makedirs(os.path.dirname(filename))
                except OSError as exception:
                    import errno

                    if exception.errno != errno.EEXIST:
                        raise

            self.file = tables.open_file(
                filename, mode=mode, filters=filters, title=title
            )
            self._opened = True
        elif isinstance(filename, tables.File):
            # filename is a pytables file
            self.file = filename
            assert self.file.isopen
            self.filename = self.file.filename
            self._opened = False
        else:
            raise TypeError(
                f"{self.__class__.__name__} must be initialised with a filename to open or "
                f"an open tables.File"
            )

        # now update metadata if given
        if metadata is not None and self.file.mode != "r":
            for k, v in metadata.items():
                setattr(self.file.root._v_attrs, k, v)

    def __del__(self):
        if self._opened and self.file.isopen:
            self.file.close()



from ..model import Model
import numpy as np
import logging

logger = logging.getLogger(__name__)


class BisectionSearchModel(Model):
    """A Pywr model that performs a bisection search.

    When a `BisectionSearchModel` is run it performs multiple simulations using a bisection
    algorithm to find the largest value of a parameter that satisfies all defined constraints. The bisection
    proceeds until a minimum gap (epsilon) is reached. After the bisection the best feasible value is re-run
    to ensure the model ends with corresponding recorder values. If no feasible solutions are found then
    an error is raised.

    Parameters
    ==========
    bisect_parameter : Parameter
        The parameter to vary during the bisection process. This parameter must have a `double_size` of 1
        and have lower and upper bounds defined; the bisection is undertaken between these bounds.
    bisect_epsilon : float
        The termination criterion for the bisection process. When the bisection has narrowed to a gap
         of less than the `bisect_epsilon` the process is terminated.
    error_on_infeasible : bool (default True)
        If true a ValueError is raised if no feasible solution is found during the bisection process. If false
        no error is raised if there is no feasible solution, and a solution using the lower bounds of the
        bisection parameter is the final result.
    """

    def __init__(self, **kwargs):
        self.bisect_parameter = kwargs.pop("bisect_parameter", None)
        self.bisect_epsilon = kwargs.pop("bisect_epsilon", None)
        self.error_on_infeasible = kwargs.pop("error_on_infeasible", True)
        super().__init__(**kwargs)

    @classmethod
    def _load_from_dict(cls, data, model=None, path=None, solver=None, **kwargs):
        model = super()._load_from_dict(
            data, model=None, path=None, solver=None, **kwargs
        )

        try:
            bisect_data = data["bisection"]
        except KeyError:
            pass
        else:
            model.bisect_parameter = bisect_data.get("parameter", None)
            model.bisect_epsilon = bisect_data.get("epsilon", None)
            model.error_on_infeasible = bisect_data.get("error_on_infeasible", True)
        return model

    def run(self):
        """Perform a bisection search using repeated simulation."""

        if self.bisect_epsilon is None:
            raise ValueError("Bisection epsilon is not defined.")
        if self.bisect_parameter is None:
            raise ValueError("Bisection parameter is not defined.")
        # Setup the model first
        self.setup()
        # Get the bisection parameter
        param = self.parameters[self.bisect_parameter]
        if param.double_size != 1 or param.integer_size != 0:
            raise ValueError(
                "Bisection is only supported using a parameter with only a single double variable "
                "(e.g ConstantParameter)."
            )
        # Use the bounds of the parameter for the bisection search space
        min_value = param.get_double_lower_bounds()[0]
        max_value = param.get_double_upper_bounds()[0]
        if min_value >= max_value:
            raise ValueError(
                "Minimum bounds of the bisection parameter must be strictly greater than its "
                "maximum bounds."
            )

        if self.bisect_epsilon <= 0.0:
            raise ValueError("Bisection epsilon value must be greater than zero.")

        logger.info(
            f'Starting bisection using parameter "{self.bisect_parameter}" with epsilon of '
            f"{self.bisect_epsilon:.4f}."
        )

        best_feasible = -np.inf
        while (max_value - min_value) > self.bisect_epsilon:
            # Compute the current value to try as the middle of the bounds
            current_value = (max_value + min_value) / 2
            logger.debug(
                f"Performing bisection run with value: {current_value:.4f}; "
                f"bounds [{min_value:.4f}, {max_value:.4f}]"
            )
            # Update parameter & perform the simulation
            param.set_double_variables(np.array([current_value]))
            super().run()
            # Check for constraint failures
            if self.is_feasible():
                # No constraint failures; tighten lower bounds
                min_value = current_value
                best_feasible = max(best_feasible, current_value)
            else:
                # Failed; tighten upper bounds
                max_value = current_value

        # Ensure a feasible solution is found.
        if np.isneginf(best_feasible):
            if self.error_on_infeasible:
                raise ValueError(
                    "No feasible solutions found during bisection. Trying lowering the bounds on the "
                    "bisection parameter."
                )
            else:
                # If no feasible the "best" value is the lower bounds.
                best_feasible = min_value
        # Finally, rerun at the best found feasible value
        # This is slightly inefficient, but ensures the model's recorders are at the best_feasible
        # value's results at the end of the simulation.
        param.set_double_variables(np.array([best_feasible]))
        ret = super().run()
        logger.info(
            f'Bisection complete! Highest feasible value of "{self.bisect_parameter}" '
            f"found: {best_feasible:.4f}"
        )
        return ret



#!/usr/bin/env python
import os
import sys
from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

if sys.platform == "win32":
    dll_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), ".libs"))
    if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
        # https://docs.python.org/3/whatsnew/3.8.html#bpo-36085-whatsnew
        if os.path.exists(dll_folder):
            os.add_dll_directory(dll_folder)
    else:
        os.environ["PATH"] = os.environ["PATH"] + ";" + dll_folder



def parameter_property(private_name):
    """Factory for parameter properties

    This property handles adding/removing parameters from the tree.

    Usage
    =====

    class NewParameter(Parameter):
        control_curve = parameter_property("_control_curve")
    """

    def parameter():
        def fget(self):
            return getattr(self, private_name)

        def fset(self, parameter):
            old_parameter = getattr(self, private_name)
            if old_parameter:
                self.children.remove(old_parameter)
            self.children.add(parameter)
            setattr(self, private_name, parameter)

        return {"fget": fget, "fset": fset}

    parameter = property(**parameter())
    return parameter



from pywr.timestepper import Timestepper  # noqa
from pywr.model import Model  # noqa
from pywr.model import (
    ModelStructureError,
    ModelDocumentWarning,
)  # noqa TODO: exceptions module?
from pywr.nodes import *  # noqa
from pywr._core import Scenario, ScenarioCollection, ScenarioIndex, Timestep  # noqa
from pywr._component import Component  # noqa
